@@ -508,7 +508,7 @@ extern char remote_port[];	/* client port */
 /*
  * public functions
  */
-extern void register_watchdog_state_change_interupt(void);
+extern void register_watchdog_state_change_interrupt(void);
 extern bool register_node_operation_request(POOL_REQUEST_KIND kind, int* node_id_set, int count, unsigned int wd_failover_id);
 extern char *get_config_file_name(void);
 extern char *get_hba_file_name(void);
@@ -612,9 +612,9 @@ extern POOL_STATUS OneNode_do_command(POOL_CONNECTION *frontend, POOL_CONNECTION
 
 /* child.c */
 extern POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
-	char *hostname, int port, char *dbname, char *user, char *password, bool retry);
+	int db_node_id, char *hostname, int port, char *dbname, char *user, char *password, bool retry);
 extern POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection_noerror(
-    char *hostname, int port, char *dbname, char *user, char *password, bool retry);
+    int db_node_id, char *hostname, int port, char *dbname, char *user, char *password, bool retry);
 extern void discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT *cp);
 
 /* define pool_system.c */
