@@ -90,6 +90,12 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *number_of_nodes);
 static POOL_STATUS close_standby_transactions(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend);
+static char *
+flatten_set_variable_args(const char *name, List *args);
+static void pool_wait_till_ready_for_query(POOL_CONNECTION_POOL *backend);
+static void pool_discard_except_sync_and_ready_for_query(POOL_CONNECTION *frontend,
+														 POOL_CONNECTION_POOL *backend);
+
 /*
  * Process Query('Q') message
  * Query messages include an SQL string.
@@ -406,7 +412,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 * - statement is INSERT
 		 * - either "INSERT LOCK" comment exists or insert_lock directive specified
 		 */
-		if (!RAW_MODE)
+		if (!RAW_MODE && !STREAM)
 		{
 			/*
 			 * If there's only one node to send the command, there's no
@@ -582,6 +588,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 	POOL_SENT_MESSAGE *bind_msg;
+	bool foundp = false;
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
@@ -622,6 +629,10 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	ereport(DEBUG2,(errmsg("Execute: query string = <%s>", query)));
 
+	ereport(DEBUG1,(errmsg("Execute: pool_is_likely_select: %d pool_is_writing_transaction: %d TSTATE: %c",
+						   pool_is_likely_select(query), pool_is_writing_transaction(),
+						   TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID))));
+
 	/*
 	 * Fetch memory cache if possible
 	 */
@@ -629,7 +640,6 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		!pool_is_writing_transaction() &&
 		(TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) != 'E'))
 	{
-		bool foundp;
 		POOL_STATUS status;
 		char *search_query = NULL;
 		int len;
@@ -638,6 +648,9 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 		len = strlen(query)+1;
 		search_query = MemoryContextStrdup(query_context->memory_context,query);
+
+		ereport(DEBUG1,(errmsg("Execute: checkig cache fetch condition")));
+		
 		/*
 		 * Add bind message's info to query to search.
 		 */
@@ -698,12 +711,26 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 		if (foundp)
 		{
+#ifdef DEBUG
+			extern bool stop_now;
+#endif
 			pool_ps_idle_display(backend);
-			pool_set_skip_reading_from_backends();
 			pool_stats_count_up_num_cache_hits();
-			pool_unset_query_in_progress();
-			return POOL_CONTINUE;
+			session_context->query_context->skip_cache_commit = true;
+#ifdef DEBUG
+			stop_now = true;
+#endif
+			
+			if (!STREAM || !pool_is_doing_extended_query_message())
+			{
+				pool_set_skip_reading_from_backends();
+				pool_stats_count_up_num_cache_hits();
+				pool_unset_query_in_progress();
+				return POOL_CONTINUE;
+			}
 		}
+		else
+			session_context->query_context->skip_cache_commit = false;
 	}
 
 	session_context->query_context = query_context;
@@ -761,8 +788,11 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	{
 		POOL_PENDING_MESSAGE *pmsg;
 
-		pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, true);
-		pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID, true);
+		if (!foundp)
+		{
+			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, true);
+			pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID, true);
+		}
 
 		/* Add pending message */
 		pmsg = pool_pending_message_create('E', len, contents);
@@ -1460,8 +1490,8 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			query_context->where_to_send[session_context->load_balance_node_id] = true;
 		}
 
-		pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID, false);
-		pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID, false);
+		pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID, true);
 
 		/* Add pending message */
 		pmsg = pool_pending_message_create('C', len, contents);
@@ -1722,101 +1752,12 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	if (pool_is_query_in_progress())
 	{
 		node = pool_get_parse_tree();
+		query = pool_get_query_string();
 
 		if (pool_is_command_success())
 		{
-			query = pool_get_query_string();
-
 			if (node)
-			{
-				/*
-				 * If the query was BEGIN/START TRANSACTION, clear the
-				 * history that we had a writing command in the transaction
-				 * and forget the transaction isolation level.
-				 *
-				 * XXX If BEGIN is received while we are already in an
-				 * explicit transaction, the command *successes*
-				 * (just with a NOTICE message). In this case we lose
-				 * "writing_transaction" etc. info.
-				 */
-				if (is_start_transaction_query(node))
-				{
-					pool_unset_writing_transaction();
-					pool_unset_failed_transaction();
-					pool_unset_transaction_isolation();
-				}
-
-				/*
-				 * If the query was COMMIT/ABORT, clear the history
-				 * that we had a writing command in the transaction
-				 * and forget the transaction isolation level.  This
-				 * is necessary if succeeding transaction is not an
-				 * explicit one.
-				 */
-				else if (is_commit_or_rollback_query(node))
-				{
-					pool_unset_writing_transaction();
-					pool_unset_failed_transaction();
-					pool_unset_transaction_isolation();
-				}
-
-				/*
-				 * SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or SET
-				 * SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL
-				 * SERIALIZABLE, remember it.
-				 */
-				else if (is_set_transaction_serializable(node))
-				{
-					pool_set_transaction_isolation(POOL_SERIALIZABLE);
-				}
-
-				/*
-				 * If 2PC commands has been executed, automatically close
-				 * transactions on standbys if there is any open
-				 * transaction since 2PC commands close transaction on
-				 * primary.
-				 */
-				else if (is_2pc_transaction_query(node))
-				{
-					if (close_standby_transactions(frontend, backend) != POOL_CONTINUE)
-						return POOL_END;
-				}
-
-				else if (!is_select_query(node, query))
-				{
-					/*
-					 * If the query was not READ SELECT, and we are in an
-					 * explicit transaction, remember that we had a write
-					 * query in this transaction.
-					 */
-					if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) == 'T')
-					{
-						/* However, if the query is "SET TRANSACTION READ ONLY" or its variant,
-						 * don't set it.
-						 */
-						if (!pool_is_transaction_read_only(node))
-						{
-							ereport(DEBUG1,
-									(errmsg("not SET TRANSACTION READ ONLY")));
-
-							pool_set_writing_transaction();
-						}
-					}
-
-					/*
-					 * If the query was CREATE TEMP TABLE, discard
-					 * temp table relcache because we might have had
-					 * persistent table relation cache which has table
-					 * name as the temp table.
-					 */
-					if (IsA(node, CreateStmt))
-					{
-						CreateStmt *create_table_stmt = (CreateStmt *)node;
-						if (create_table_stmt->relation->relpersistence)
-							discard_temp_table_relcache();
-					}
-				}
-			}
+				pool_at_command_success(frontend, backend);
 
 			/* Memory cache enabled? */
 			if (cache_commit && pool_config->memory_cache_enabled)
@@ -1860,7 +1801,8 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			if ((pool_is_doing_extended_query_message() &&
 				 session_context->query_context->query_state[MASTER_NODE_ID] != POOL_UNPARSED &&
 			     session_context->uncompleted_message) ||
-			    (!pool_is_doing_extended_query_message() && session_context->uncompleted_message))
+			    (!pool_is_doing_extended_query_message() && session_context->uncompleted_message &&
+				 session_context->uncompleted_message->kind != 0))
 			{
 				pool_add_sent_message(session_context->uncompleted_message);
 				pool_remove_sent_message(session_context->uncompleted_message->kind,
@@ -1978,12 +1920,22 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 	{
 		POOL_PENDING_MESSAGE *pmsg;
 
-		pmsg = pool_pending_message_get(POOL_CLOSE);
+		pmsg = pool_pending_message_pull_out();
 
 		if (pmsg)
 		{
-			kind = pool_get_close_message_spec(pmsg);
-			name = pool_get_close_message_name(pmsg);
+			/* Sanity check */
+			if (pmsg->type != POOL_CLOSE)
+			{
+				ereport(LOG,
+						(errmsg("CloseComplete: pending messge was not Close request: %s", pool_pending_message_type_to_string(pmsg->type))));
+			}
+			else
+			{
+				kind = pool_get_close_message_spec(pmsg);
+				name = pool_get_close_message_name(pmsg);
+				kind = kind=='S'?'P':'B';
+			}
 			pool_pending_message_free_pending_message(pmsg);
 		}
 	}
@@ -2002,12 +1954,16 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 		}
 	}
 	
-	if (kind == ' ')
+	if (kind != ' ')
 	{
 		pool_remove_sent_message(kind, name);
 		ereport(DEBUG1,
-				(errmsg("CloseComplete: remove uncompleted message. kind:%c, name:%s",
+				(errmsg("CloseComplete: remove sent message. kind:%c, name:%s",
 						kind, name)));
+		if (pool_config->memory_cache_enabled)
+		{
+			pool_discard_temp_query_cache(pool_get_current_cache());
+		}
 	}
 
 	return status;
@@ -2102,8 +2058,9 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 	if (ret != POOL_CONTINUE)
 		return ret;
 
-	raise_intentional_error_if_need(backend);
-	
+	if (!STREAM)
+		raise_intentional_error_if_need(backend);
+
 	return POOL_CONTINUE;
 }
 
@@ -2352,6 +2309,12 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			else if (!pool_is_query_in_progress())
 				pool_set_query_in_progress();
 			status = SimpleForwardToBackend(fkind, frontend, backend, len, contents);
+
+			if (STREAM)
+			{
+				/* Wait till Ready for query received */
+				pool_wait_till_ready_for_query(backend);
+			}
 			break;
 
 		case 'F':	/* FunctionCall */
@@ -2432,9 +2395,9 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 			pool_unset_query_in_progress();
 
 		/*
-		 * Check if If we have pending data in backend connection cache. If we
-		 * do, it is likely that a sync message has been sent to backend and the
-		 * backend replied back to us. So we need to process it.
+		 * Check if we have pending data in backend connection cache. If we
+		 * do, it is likely that a sync message has been sent to backend and
+		 * the backend replied back to us. So we need to process it.
 		 */
 		if (is_backend_cache_empty(backend))
 		{
@@ -2482,6 +2445,11 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 					(errmsg("processing backend response"),
 						 errdetail("Ready For Query received")));
 				status = ReadyForQuery(frontend, backend, true, true);
+#ifdef DEBUG
+				extern bool stop_now;
+				if (stop_now)
+					exit(0);
+#endif
 				break;
 
 			case '1':	/* ParseComplete */
@@ -2505,15 +2473,13 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 				}
 				status = ParseComplete(frontend, backend);
 				pool_set_command_success();
-				if (STREAM||REPLICATION||RAW_MODE)
-					pool_unset_query_in_progress();
+				pool_unset_query_in_progress();
 				break;
 
 			case '2':	/* BindComplete */
 				status = BindComplete(frontend, backend);
 				pool_set_command_success();
-				if (STREAM||REPLICATION||RAW_MODE)
-					pool_unset_query_in_progress();
+				pool_unset_query_in_progress();
 				break;
 
 			case '3':	/* CloseComplete */
@@ -2525,10 +2491,17 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 					{
 						/* parse_before_bind() was called. Do not foward the
 						 * close complete message to frontend. */
-						ereport(LOG,
+						ereport(DEBUG1,
 								(errmsg("processing backend response"),
 								 errdetail("do not forward close complete message to frontend")));
 						pool_discard_packet_contents(backend);
+						
+						/* Remove pending message here because
+						 * read_kind_from_backend() did not do it.
+						 */
+						pmsg = pool_pending_message_pull_out();
+						pool_pending_message_free_pending_message(pmsg);
+
 						pool_unset_query_in_progress();
 						pool_set_command_success();
 						status = POOL_CONTINUE;
@@ -2538,8 +2511,7 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 				status = CloseComplete(frontend, backend);
 				pool_set_command_success();
-				if (STREAM||REPLICATION||RAW_MODE)
-					pool_unset_query_in_progress();
+				pool_unset_query_in_progress();
 				break;
 
 			case 'E':	/* ErrorResponse */
@@ -2552,31 +2524,8 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 				{
 					pool_set_ignore_till_sync();
 					pool_unset_query_in_progress();
-
 					if (STREAM)
-					{
-						POOL_PENDING_MESSAGE *pmsg;
-						int i;
-
-						/* Remove all pending messages */
-						do
-						{
-							pmsg = pool_pending_message_pull_out();
-							pool_pending_message_free_pending_message(pmsg);
-						}
-						while (pmsg);
-
-						pool_pending_message_reset_previous_message();
-
-						/* Discard read buffer */
-						for (i=0;i<NUM_BACKENDS;i++)
-						{
-							if (VALID_BACKEND(i))
-							{
-								pool_discard_read_buffer(CONNECTION(backend, i));
-							}
-						}
-					}
+						pool_discard_except_sync_and_ready_for_query(frontend, backend);
 				}
 				break;
 
@@ -3419,3 +3368,285 @@ flatten_set_variable_args(const char *name, List *args)
 	
 	return buf.data;
 }
+
+/* Called when sync message is received.
+ * Wait till ready for query received.
+ */
+static void pool_wait_till_ready_for_query(POOL_CONNECTION_POOL *backend)
+{
+	char kind;
+	int len;
+	int poplen;
+	char *buf;
+	int i;
+
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (VALID_BACKEND(i))
+		{
+			for (;;)
+			{
+				pool_read(CONNECTION(backend, i), &kind, sizeof(kind));
+				ereport(DEBUG1,
+						(errmsg("pool_wait_till_ready_for_query: kind: %c", kind)));
+				pool_push(CONNECTION(backend, i), &kind, sizeof(kind));
+				pool_read(CONNECTION(backend, i), &len, sizeof(len));
+				pool_push(CONNECTION(backend, i), &len, sizeof(len));
+				if ((ntohl(len)-sizeof(len)) > 0)
+				{
+					buf = pool_read2(CONNECTION(backend, i), ntohl(len)-sizeof(len));
+					pool_push(CONNECTION(backend, i), buf, ntohl(len)-sizeof(len));
+				}
+
+				if (kind == 'Z')	/* Ready for query? */
+				{
+					pool_pop(CONNECTION(backend, i), &poplen);
+					ereport(DEBUG1,
+							(errmsg("pool_wait_till_ready_for_query: backend:%d ready for query found. buffer length:%d",
+									i, CONNECTION(backend, i)->len)));
+					break;
+				}
+			}
+		}
+	}
+}
+
+/*
+ * Called when error response received in streaming replication mode and doing
+ * extended query. Remove all pending messages and backend message buffer data
+ * except POOL_SYNC pending message and ready for query.  If sync message is
+ * not received yet, call ProcessFrontendResponse() to read data from
+ * frontend.
+ */
+static void pool_discard_except_sync_and_ready_for_query(POOL_CONNECTION *frontend,
+														 POOL_CONNECTION_POOL *backend)
+{
+	POOL_PENDING_MESSAGE *pmsg;
+	int i;
+
+	if (!pool_is_doing_extended_query_message() || !STREAM)
+		return;
+
+	/*
+	 * Check to see if we aready received a sync
+	 * message. If not, call ProcessFrontendResponse() to
+	 * get the sync message from client.
+	 */
+	pmsg = pool_pending_message_get(POOL_SYNC);
+	if (pmsg == NULL)
+	{
+		char kind;
+		int len;
+		POOL_PENDING_MESSAGE *msg;
+		char *contents;
+
+		for(;;)
+		{
+			pool_read(frontend, &kind, sizeof(kind));
+			pool_read(frontend, &len, sizeof(len));
+			len = ntohl(len) - sizeof(len);
+			if (len > 0)
+				contents = pool_read2(frontend, len);
+			if (kind == 'S')
+			{
+				msg = pool_pending_message_create('S', 0, NULL);
+				pool_pending_message_add(msg);
+				pool_pending_message_free_pending_message(msg);
+				SimpleForwardToBackend(kind, frontend, backend, len, contents);
+				break;
+			}
+		}
+	}
+	else
+		pool_pending_message_free_pending_message(pmsg);
+
+	/* Remove all pending messages except sync message */
+	do
+	{
+		pmsg = pool_pending_message_head_message();
+		if (pmsg && pmsg->type == POOL_SYNC)
+		{
+			ereport(LOG,
+					(errmsg("Process backend response: sync pending message found after receiving error response")));
+			pool_unset_ignore_till_sync();
+			pool_pending_message_free_pending_message(pmsg);
+			break;
+		}
+		pool_pending_message_free_pending_message(pmsg);
+		pmsg = pool_pending_message_pull_out();
+		pool_pending_message_free_pending_message(pmsg);
+	}
+	while (pmsg);
+
+	pool_pending_message_reset_previous_message();
+
+	/* Discard read buffer execpt "Ready for query" */
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (VALID_BACKEND(i))
+		{
+			char kind;
+			int len;
+			int sts;
+
+			while (!pool_read_buffer_is_empty(CONNECTION(backend, i)))
+			{
+				sts = pool_read(CONNECTION(backend, i), &kind, sizeof(kind));
+				if (sts < 0 || kind == '\0')
+				{
+					ereport(DEBUG1,
+							(errmsg("pool_discard_except_sync_and_ready_for_query: EOF detected while reading from backend: %d buffer length: %d sts: %d",
+									i, CONNECTION(backend, i)->len, sts)));
+					pool_unread(CONNECTION(backend, i), &kind, sizeof(kind));
+					break;
+				}
+
+				if (kind == 'Z')	/* Ready for query? */
+				{
+					pool_unread(CONNECTION(backend, i), &kind, sizeof(kind));
+					ereport(DEBUG1,
+							(errmsg("pool_discard_except_sync_and_ready_for_query: Ready for query found. backend:%d",
+									i)));
+					break;
+				}
+				else
+				{
+					/* Read and discard packet */
+					pool_read(CONNECTION(backend, i), &len, sizeof(len));
+					if ((ntohl(len)-sizeof(len)) > 0)
+					{
+						pool_read2(CONNECTION(backend, i), ntohl(len)-sizeof(len));
+					}
+					ereport(DEBUG1,
+							(errmsg("pool_discard_except_sync_and_ready_for_query: discarding packet %c (len:%lu) of backend:%d", kind, ntohl(len)-sizeof(len), i)));
+				}
+			}
+		}
+	}
+}
+
+/*
+ * Handle misc treatment when a command successfully completed.
+ * Preconditions: query is in progress. The command is succeeded.
+ */
+void pool_at_command_success(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
+{
+	Node *node;
+	char *query;
+
+	/* Sanity checks */
+	if (!pool_is_query_in_progress())
+	{
+		ereport(ERROR,
+				(errmsg("pool_at_command_success: query is not in progress")));
+	}
+
+	if (!pool_is_command_success())
+	{
+		ereport(ERROR,
+				(errmsg("pool_at_command_success: command did not suceed")));
+	}
+
+	node = pool_get_parse_tree();
+
+	if (!node)
+	{
+		ereport(ERROR,
+				(errmsg("pool_at_command_success: no parse tree found")));
+	}
+
+	query = pool_get_query_string();
+	
+	if (query == NULL)
+	{
+		ereport(ERROR,
+				(errmsg("pool_at_command_success: no query found")));
+	}
+
+	/*
+	 * If the query was BEGIN/START TRANSACTION, clear the
+	 * history that we had a writing command in the transaction
+	 * and forget the transaction isolation level.
+	 *
+	 * XXX If BEGIN is received while we are already in an
+	 * explicit transaction, the command *successes*
+	 * (just with a NOTICE message). In this case we lose
+	 * "writing_transaction" etc. info.
+	 */
+	if (is_start_transaction_query(node))
+	{
+		pool_unset_writing_transaction();
+		pool_unset_failed_transaction();
+		pool_unset_transaction_isolation();
+	}
+
+	/*
+	 * If the query was COMMIT/ABORT, clear the history
+	 * that we had a writing command in the transaction
+	 * and forget the transaction isolation level.  This
+	 * is necessary if succeeding transaction is not an
+	 * explicit one.
+	 */
+	else if (is_commit_or_rollback_query(node))
+	{
+		pool_unset_writing_transaction();
+		pool_unset_failed_transaction();
+		pool_unset_transaction_isolation();
+	}
+
+	/*
+	 * SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or SET
+	 * SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL
+	 * SERIALIZABLE, remember it.
+	 */
+	else if (is_set_transaction_serializable(node))
+	{
+		pool_set_transaction_isolation(POOL_SERIALIZABLE);
+	}
+
+	/*
+	 * If 2PC commands has been executed, automatically close
+	 * transactions on standbys if there is any open
+	 * transaction since 2PC commands close transaction on
+	 * primary.
+	 */
+	else if (is_2pc_transaction_query(node))
+	{
+		close_standby_transactions(frontend, backend);
+	}
+
+	else if (!is_select_query(node, query))
+	{
+		/*
+		 * If the query was not READ SELECT, and we are in an
+		 * explicit transaction, remember that we had a write
+		 * query in this transaction.
+		 */
+		if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) == 'T')
+		{
+			/* However, if the query is "SET TRANSACTION READ ONLY" or its variant,
+			 * don't set it.
+			 */
+			if (!pool_is_transaction_read_only(node))
+			{
+				ereport(DEBUG1,
+						(errmsg("not SET TRANSACTION READ ONLY")));
+
+				pool_set_writing_transaction();
+			}
+		}
+
+		/*
+		 * If the query was CREATE TEMP TABLE, discard
+		 * temp table relcache because we might have had
+		 * persistent table relation cache which has table
+		 * name as the temp table.
+		 */
+		if (IsA(node, CreateStmt))
+		{
+			CreateStmt *create_table_stmt = (CreateStmt *)node;
+			if (create_table_stmt->relation->relpersistence == 't')
+				discard_temp_table_relcache();
+		}
+	}
+}
