@@ -38,11 +38,11 @@
 typedef struct {
 	char	*attrname;	/* attribute name */
 	char	*adsrc;		/* default value expression */
-	int		 use_timestamp;
+	int		 use_timestamp;	/* not zero if timestamp is used in default value */
 } TSAttr;
 
 typedef struct {
-	int		relnatts;
+	int		relnatts;	/* num of attributes */
 	TSAttr	attr[1];
 } TSRel;
 
@@ -51,7 +51,7 @@ typedef struct {
 	POOL_CONNECTION_POOL	*backend;
 	char					*relname;
 	int						 num_params;		/* num of original params (for Parse) */
-	bool		 			 rewrite_to_params;	/* true if rewritten to param insread of const */
+	bool		 			 rewrite_to_params;	/* true if timestamp is rewritten to param insread of const */
 	bool		 			 rewrite;			/* has rewritten? */
 	List					*params;			/* list of additional params */
 } TSRewriteContext;
@@ -89,9 +89,11 @@ ts_register_func(POOL_SELECT_RESULT *res)
 	{
 		int index = 0;
 
+		/* attribute name */
 		rel->attr[i].attrname = strdup(res->data[i * NUM_COLS + index]);
 		index++;
 
+		/* attribute default value */
 		if (res->data[i * NUM_COLS + index])
 			rel->attr[i].adsrc = strdup(res->data[i * NUM_COLS + index]);
 		else
@@ -99,6 +101,7 @@ ts_register_func(POOL_SELECT_RESULT *res)
 
 		index++;
 
+		/* if timestamp is used in default value */
 		rel->attr[i].use_timestamp = *(res->data[i * NUM_COLS + index]) == 't';
 		ereport(DEBUG1,
 			(errmsg("timestamp register function"),
@@ -107,6 +110,7 @@ ts_register_func(POOL_SELECT_RESULT *res)
 					   rel->attr[i].use_timestamp)));
 	}
 
+	/* num of attributes */
 	rel->relnatts = res->numrows;
 	return (void *) rel;
 }
@@ -125,10 +129,7 @@ ts_unregister_func(void *data)
 static TSRel*
 relcache_lookup(TSRewriteContext *ctx)
 {
-#define ATTRDEFQUERY (Pgversion(ctx->backend)->major >= 73 ? \
-	"SELECT attname, pg_get_expr(d.adbin, d.adrelid), coalesce((pg_get_expr(d.adbin, d.adrelid) LIKE '%%now()%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%''now''::text%%' OR" \
-	" pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_TIMESTAMP%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_TIME%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_DATE%%' OR" \
-	" pg_get_expr(d.adbin, d.adrelid) LIKE '%%LOCALTIME%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%LOCALTIMESTAMP%%')" \
+#define ATTRDEFQUERY "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
 	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
 	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
 	" a.atttypid = 'date'::regtype::oid OR" \
@@ -138,25 +139,9 @@ relcache_lookup(TSRewriteContext *ctx)
 	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
 	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
 	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s'" \
-	" ORDER BY a.attnum" : \
-	"SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%' OR" \
-	" d.adsrc LIKE '%%CURRENT_TIMESTAMP%%' OR d.adsrc LIKE '%%CURRENT_TIME%%' OR d.adsrc LIKE '%%CURRENT_DATE%%' OR" \
-	" d.adsrc LIKE '%%LOCALTIME%%' OR d.adsrc LIKE '%%LOCALTIMESTAMP%%')" \
-	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
-	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
-	" a.atttypid = 'date'::regtype::oid OR" \
-	" a.atttypid = 'time'::regtype::oid OR" \
-	" a.atttypid = 'time with time zone'::regtype::oid)" \
-    " , false)" \
-	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
-	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
-	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s'" \
-	" ORDER BY a.attnum")
+	" ORDER BY a.attnum"
 
-#define ATTRDEFQUERY2 (Pgversion(ctx->backend)->major >= 73 ? \
-	"SELECT attname, pg_get_expr(d.adbin, d.adrelid), coalesce((pg_get_expr(d.adbin, d.adrelid) LIKE '%%now()%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%''now''::text%%' OR" \
-	" pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_TIMESTAMP%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_TIME%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_DATE%%' OR" \
-	" pg_get_expr(d.adbin, d.adrelid) LIKE '%%LOCALTIME%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%LOCALTIMESTAMP%%')" \
+#define ATTRDEFQUERY2 "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
 	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
 	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
 	" a.atttypid = 'date'::regtype::oid OR" \
@@ -166,25 +151,9 @@ relcache_lookup(TSRewriteContext *ctx)
 	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
 	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
 	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = pgpool_regclass('%s')" \
-	" ORDER BY a.attnum" : \
-	"SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%' OR" \
-	" d.adsrc LIKE '%%CURRENT_TIMESTAMP%%' OR d.adsrc LIKE '%%CURRENT_TIME%%' OR d.adsrc LIKE '%%CURRENT_DATE%%' OR" \
-	" d.adsrc LIKE '%%LOCALTIME%%' OR d.adsrc LIKE '%%LOCALTIMESTAMP%%')" \
-	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
-	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
-	" a.atttypid = 'date'::regtype::oid OR" \
-	" a.atttypid = 'time'::regtype::oid OR" \
-	" a.atttypid = 'time with time zone'::regtype::oid)" \
-    " , false)" \
-	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
-	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
-	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = pgpool_regclass('%s')" \
-	" ORDER BY a.attnum")
+	" ORDER BY a.attnum"
 
-#define ATTRDEFQUERY3 (Pgversion(ctx->backend)->major >= 73 ? \
-	"SELECT attname, pg_get_expr(d.adbin, d.adrelid), coalesce((pg_get_expr(d.adbin, d.adrelid) LIKE '%%now()%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%''now''::text%%' OR" \
-	" pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_TIMESTAMP%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_TIME%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%CURRENT_DATE%%' OR" \
-	" pg_get_expr(d.adbin, d.adrelid) LIKE '%%LOCALTIME%%' OR pg_get_expr(d.adbin, d.adrelid) LIKE '%%LOCALTIMESTAMP%%')" \
+#define ATTRDEFQUERY3 "SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%')" \
 	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
 	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
 	" a.atttypid = 'date'::regtype::oid OR" \
@@ -194,20 +163,7 @@ relcache_lookup(TSRewriteContext *ctx)
 	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
 	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
 	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = to_regclass('%s')" \
-	" ORDER BY a.attnum" : \
-	"SELECT attname, d.adsrc, coalesce((d.adsrc LIKE '%%now()%%' OR d.adsrc LIKE '%%''now''::text%%' OR" \
-	" d.adsrc LIKE '%%CURRENT_TIMESTAMP%%' OR d.adsrc LIKE '%%CURRENT_TIME%%' OR d.adsrc LIKE '%%CURRENT_DATE%%' OR" \
-	" d.adsrc LIKE '%%LOCALTIME%%' OR d.adsrc LIKE '%%LOCALTIMESTAMP%%')" \
-	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
-	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
-	" a.atttypid = 'date'::regtype::oid OR" \
-	" a.atttypid = 'time'::regtype::oid OR" \
-	" a.atttypid = 'time with time zone'::regtype::oid)" \
-    " , false)" \
-	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
-	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
-	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = to_regclass('%s')" \
-	" ORDER BY a.attnum")
+	" ORDER BY a.attnum"
 
 	char *query;
 	char *table_name;
@@ -243,6 +199,10 @@ relcache_lookup(TSRewriteContext *ctx)
 	return (TSRel *) pool_search_relcache(ts_relcache, ctx->backend, table_name);
 }
 
+/* If timestamp valus is required then return ts_const. Otherwize make
+ * a new ParamRef, whose number member is 0 and is overwritten at last,
+ * and add it into params list in context.
+ */
 static Node *
 makeTsExpr(TSRewriteContext *ctx)
 {
@@ -318,55 +278,89 @@ rewrite_timestamp_walker(Node *node, void *context)
 	if (node == NULL)
 		return false;
 
-	if (nodeTag(node) == T_FuncCall)
+	switch (nodeTag(node))
 	{
-		/* `now()' FuncCall */
-		FuncCall	*fcall = (FuncCall *) node;
-
-		if ((list_length(fcall->funcname) == 1 &&
-			 strcmp("now", strVal(linitial(fcall->funcname))) == 0) ||
-			(list_length(fcall->funcname) == 2 &&
-			 strcmp("pg_catalog", strVal(linitial(fcall->funcname))) == 0 &&
-			 strcmp("now", strVal(lsecond(fcall->funcname))) == 0))
-		{
-			TypeCast	*tc = makeNode(TypeCast);
-			tc->arg = makeTsExpr(ctx);
-			tc->typeName = SystemTypeName("text");
-
-			fcall->funcname = SystemFuncName("timestamptz");
-			fcall->args = list_make1(tc);
-			ctx->rewrite = true;
-		}
-	}
-	else if (IsA(node, TypeCast))
-	{
-		/* CURRENT_DATE, CURRENT_TIME, LOCALTIMESTAMP, LOCALTIME etc.*/
-		TypeCast	*tc = (TypeCast *) node;
-
-		if ((isSystemType((Node *) tc->typeName, "date") ||
-			 isSystemType((Node *) tc->typeName, "timestamp") ||
-			 isSystemType((Node *) tc->typeName, "timestamptz") ||
-			 isSystemType((Node *) tc->typeName, "time") ||
-			 isSystemType((Node *) tc->typeName, "timetz")))
-		{
-			/* rewrite `'now'::timestamp' and `'now'::text::timestamp' both */
-			if (isSystemTypeCast(tc->arg, "text"))
-				tc = (TypeCast *) tc->arg;
-
-			if (isStringConst(tc->arg, "now"))
+		case T_InsertStmt:
 			{
-				tc->arg = (Node *) makeTsExpr(ctx);
-				ctx->rewrite = true;
-			}
-		}
-	}
-	else if (IsA(node, ParamRef))
-	{
-		ParamRef	*param = (ParamRef *) node;
+				TSRewriteContext	*ctx = (TSRewriteContext *) context;
+				char *relname_old;
+				InsertStmt *i_stmt = (InsertStmt *) node;
 
-		/* count how many params in original query */
-		if (ctx->num_params < param->number)
-			ctx->num_params = param->number;
+				relname_old = ctx->relname;
+				ctx->relname = make_table_name_from_rangevar(i_stmt->relation);
+				rewrite_timestamp_insert(i_stmt, ctx);
+				ctx->relname = relname_old;
+			}
+			/* tree walker is called in rewrite_timestamp_insert() */ 
+			return false;
+		case T_UpdateStmt:
+			{
+				TSRewriteContext	*ctx = (TSRewriteContext *) context;
+				char *relname_old;
+				UpdateStmt *u_stmt = (UpdateStmt *) node;
+
+				relname_old = ctx->relname;
+				ctx->relname = make_table_name_from_rangevar(u_stmt->relation);
+				rewrite_timestamp_update(u_stmt, ctx);
+				ctx->relname = relname_old;
+			}
+			/* tree walker is called in rewrite_timestamp_update() */ 
+			return false;
+		case T_FuncCall:
+			{
+				/* `now()' FuncCall */
+				FuncCall	*fcall = (FuncCall *) node;
+
+				if ((list_length(fcall->funcname) == 1 &&
+			 		strcmp("now", strVal(linitial(fcall->funcname))) == 0) ||
+					(list_length(fcall->funcname) == 2 &&
+			 		strcmp("pg_catalog", strVal(linitial(fcall->funcname))) == 0 &&
+			 		strcmp("now", strVal(lsecond(fcall->funcname))) == 0))
+				{
+					TypeCast	*tc = makeNode(TypeCast);
+					tc->arg = makeTsExpr(ctx);
+					tc->typeName = SystemTypeName("text");
+	
+					fcall->funcname = SystemFuncName("timestamptz");
+					fcall->args = list_make1(tc);
+					ctx->rewrite = true;
+				}
+			}
+			break;
+		case T_TypeCast:
+			{
+				/* CURRENT_DATE, CURRENT_TIME, LOCALTIMESTAMP, LOCALTIME etc.*/
+				TypeCast	*tc = (TypeCast *) node;
+
+				if ((isSystemType((Node *) tc->typeName, "date") ||
+			 		isSystemType((Node *) tc->typeName, "timestamp") ||
+			 		isSystemType((Node *) tc->typeName, "timestamptz") ||
+			 		isSystemType((Node *) tc->typeName, "time") ||
+			 		isSystemType((Node *) tc->typeName, "timetz")))
+				{
+					/* rewrite `'now'::timestamp' and `'now'::text::timestamp' both */
+					if (isSystemTypeCast(tc->arg, "text"))
+						tc = (TypeCast *) tc->arg;
+
+					if (isStringConst(tc->arg, "now"))
+					{
+						tc->arg = (Node *) makeTsExpr(ctx);
+						ctx->rewrite = true;
+					}
+				}
+			}
+			break;
+		case T_ParamRef:
+			{
+				ParamRef	*param = (ParamRef *) node;
+
+				/* count how many params in original query */
+				if (ctx->num_params < param->number)
+					ctx->num_params = param->number;
+			}
+			break;
+		default:
+			break;
 	}
 
 	return raw_expression_tree_walker(node, rewrite_timestamp_walker, context);
@@ -407,7 +401,25 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 	bool	 rewrite = false;
 	TSRel	*relcache;
 
+	raw_expression_tree_walker(
+			(Node *) i_stmt->withClause,
+			rewrite_timestamp_walker, (void *) ctx);
 
+	raw_expression_tree_walker(
+			(Node *) i_stmt->onConflictClause,
+			rewrite_timestamp_walker, (void *) ctx);
+
+	raw_expression_tree_walker(
+			(Node *) i_stmt->returningList,
+			rewrite_timestamp_walker, (void *) ctx);
+
+	rewrite = ctx->rewrite;
+
+	/*
+	 * INSERT INTO rel DEFAULT VALUES
+	 * rewrite to:
+	 * INSERT INTO rel VALUES (DEFAULT, '2009-..',...)
+	 */
 	if (i_stmt->selectStmt == NULL)
 	{
 		List		*values = NIL;
@@ -417,12 +429,6 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 		if (relcache == NULL)
 			return false;
 
-		/*
-		 * INSERT INTO rel DEFAULT VALUES
-		 * rewrite to:
-		 * INSERT INTO rel VALUES (DEFAULT, '2009-..',...)
-		 */
-
 		for (i = 0; i < relcache->relnatts; i++)
 		{
 			if (relcache->attr[i].use_timestamp)
@@ -442,8 +448,6 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			selectStmt->valuesLists = list_make1(values);
 			i_stmt->selectStmt = (Node *) selectStmt;
 		}
-
-		return rewrite;
 	}
 	else if (IsA(i_stmt->selectStmt, SelectStmt))
 	{
@@ -528,11 +532,11 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			 * if timestamp column is not given by column list
 			 * add colname to column list and add timestamp to values list.
 			 */
-			int			append_columns = 0;
-			int			*append_columns_list;
+			int			appended_columns = 0;
+			int			*appended_columns_list;
 			ResTarget	*col;
 
-			append_columns_list = (int *)malloc(sizeof(int)*relcache->relnatts);
+			appended_columns_list = (int *)malloc(sizeof(int)*relcache->relnatts);
 
 			for (i = 0; i < relcache->relnatts; i++)
 			{
@@ -547,16 +551,16 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 						break;
 				}
 
+				/* If columns are not found in query, append these.*/
 				if (lc_col == NULL)
 				{
-					/* column not found in query, append it.*/
 					rewrite = true;
 					col = makeNode(ResTarget);
 					col->name = relcache->attr[i].attrname;
 					col->indirection = NIL;
 					col->val = NULL;
 					i_stmt->cols = lappend(i_stmt->cols, col);
-					append_columns_list[append_columns++] = i;
+					appended_columns_list[appended_columns++] = i;
 				}
 			}
 
@@ -564,7 +568,7 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 			{
 				List		*values = lfirst(lc_row);
 
-				/* replace DEFAULT to literal */
+				/* replace DEFAULT in VALUES to literal */
 				forboth (lc_col, i_stmt->cols, lc_val, values)
 				{
 					col = lfirst(lc_col);
@@ -584,17 +588,17 @@ rewrite_timestamp_insert(InsertStmt *i_stmt, TSRewriteContext *ctx)
 					}
 				}
 
-				/* add ts_const to values list */
-				for (i = 0; i < append_columns; i++)
+				/* add timestamp value to values list */
+				for (i = 0; i < appended_columns; i++)
 				{
 					if (ctx->rewrite_to_params)
 						values = lappend(values, makeTsExpr(ctx));
 					else
 						values = lappend(values,
-										 makeStringConstFromQuery(ctx->backend, relcache->attr[append_columns_list[i]].adsrc));
+										 makeStringConstFromQuery(ctx->backend, relcache->attr[appended_columns_list[i]].adsrc));
 				}
 			}
-			free(append_columns_list);
+			free(appended_columns_list);
 		}
 	}
 
@@ -625,6 +629,14 @@ rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx)
 			(Node *) u_stmt->fromClause,
 			rewrite_timestamp_walker, (void *) ctx);
 
+	raw_expression_tree_walker(
+			(Node *) u_stmt->withClause,
+			rewrite_timestamp_walker, (void *) ctx);
+
+	raw_expression_tree_walker(
+			(Node *) u_stmt->returningList,
+			rewrite_timestamp_walker, (void *) ctx);
+
 	rewrite = ctx->rewrite;
 
 	/* rewrite "update ... set col1 = default" */
@@ -661,6 +673,8 @@ rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx)
 
 /*
  * Rewrite `now()' to timestamp literal.
+ * If rewrite_to_params is false then, we rewrite `now()' to timestamp constant.
+ * Otherwize, we rewrite `now()' to params and expand that at Bind message.
  * returns query string as palloced string, or NULL if not to need rewrite.
  */
 char *
@@ -699,16 +713,17 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 	else
 		stmt = node;
 
+
 	if (IsA(stmt, InsertStmt))
 	{
 		InsertStmt *i_stmt = (InsertStmt *) stmt;
-		ctx.relname = nodeToString(i_stmt->relation);
+		ctx.relname = make_table_name_from_rangevar(i_stmt->relation);
 		rewrite = rewrite_timestamp_insert(i_stmt, &ctx);
 	}
 	else if (IsA(stmt, UpdateStmt))
 	{
 		UpdateStmt *u_stmt = (UpdateStmt *) stmt;
-		ctx.relname = nodeToString(u_stmt->relation);
+		ctx.relname = make_table_name_from_rangevar(u_stmt->relation);
 		rewrite = rewrite_timestamp_update(u_stmt, &ctx);
 	}
 	else if (IsA(stmt, DeleteStmt))
@@ -721,6 +736,15 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 		raw_expression_tree_walker(
 				(Node *) d_stmt->whereClause,
 				rewrite_timestamp_walker, (void *) &ctx);
+
+		raw_expression_tree_walker(
+				(Node *) d_stmt->returningList,
+				rewrite_timestamp_walker, (void *) &ctx);
+
+		raw_expression_tree_walker(
+				(Node *) d_stmt->withClause,
+				rewrite_timestamp_walker, (void *) &ctx);
+
 		rewrite = ctx.rewrite;
 	}
 	else if (IsA(stmt, ExecuteStmt))
@@ -734,7 +758,9 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 
 		rewrite = ctx.rewrite;
 
-		/* add params */
+		/* add timestamp value into params list.
+		 * The number of params is stored in message
+		 */
 		if (message)
 		{
 			int		i;
@@ -746,49 +772,22 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 			}
 		}
 	}
-	else if (IsA(stmt, CreateTableAsStmt))
-	{
-		CreateTableAsStmt *c_stmt = (CreateTableAsStmt *) stmt;
-
-		/*
-		 * CREATE TABLE t1 AS SELECT now();
-		 */
-		if (IsA(c_stmt->query, SelectStmt) && c_stmt->relkind == OBJECT_TABLE)
-		{
-			/* rewrite params */
-			raw_expression_tree_walker(
-									   (Node *) c_stmt->query,
-									   rewrite_timestamp_walker, (void *) &ctx);
-
-			rewrite = ctx.rewrite;
-		}
-	}
-	else if (IsA(stmt, SelectStmt))
-	{
-		SelectStmt *s_stmt = (SelectStmt *) stmt;
-
-		/*
-		 * SELECT now() INTO t1;
-		 */
-		if (s_stmt->intoClause)
-		{
-			raw_expression_tree_walker(
-									   (Node *) s_stmt,
-									   rewrite_timestamp_walker, (void *) &ctx);
-		}
-
-		rewrite = ctx.rewrite;
-	}
 	else
 		;
 
-	/* save number of parameters in original query */
+	/* PREPARE, EXECUTE or Parse:
+	 * save the number of parameters in the original query
+	 */
 	if (message)
 		message->query_context->num_original_params = ctx.num_params;
 
+	/* don't rewrite the query if not necessary */
 	if (!rewrite)
 		return NULL;
 
+	/* PREPARE or Parse:
+	 * handle additinal parameters for timestamps
+	 */
 	if (ctx.rewrite_to_params && message)
 	{
 		ListCell	*lc;
@@ -801,10 +800,10 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 			param->number = num++;
 		}
 
-		/* save to portal */
+		/* save the number of additional params to portal */
 		message->num_tsparams = list_length(ctx.params);
 
-		/* add param type */
+		/* add type names of timestamptz into argtypes */
 		if (IsA(node, PrepareStmt))
 		{
 			int				 i;
@@ -834,7 +833,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 
 
 /*
- * rewrite Bind message to add parameter
+ * rewrite Bind message to add parameter values
  */
 char *
 bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
@@ -843,8 +842,7 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 {
 	int16		 tmp2,
 				 num_params,
-				 num_formats,
-				 num_formats_new;
+				 num_formats;
 	int32		 tmp4;
 	int			 i,
 				 ts_len,
@@ -857,11 +855,10 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 
 #ifdef TIMESTAMPDEBUG
 	fprintf(stderr, "message length:%d\n", *len);
-	for(i = 0; i < *len; i++)
+	for(i=0;i<*len;i++)
 	{
 		fprintf(stderr, "%02x ", orig_msg[i]);
 	}
-	fprintf(stderr, "\n");
 #endif
 
 	ts = get_current_timestamp(backend);
@@ -879,78 +876,53 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	/* allocate extra memory for parameter formats */
 	num_org_params = message->query_context->num_original_params;
 	new_msg = copy_to = (char *) palloc(*len + sizeof(int16) * (message->num_tsparams + num_org_params));
-
 	copy_from = orig_msg;
 
-	/* 1. portal_name */
+	/* portal_name */
 	copy_len = strlen(copy_from) + 1;
 	memcpy(copy_to, copy_from, copy_len);
-	copy_to += copy_len;
-	copy_from += copy_len;
+	copy_to += copy_len; copy_from += copy_len;
 
-	/* 2. stmt_name */
+	/* stmt_name */
 	copy_len = strlen(copy_from) + 1;
 	memcpy(copy_to, copy_from, copy_len);
-	copy_to += copy_len;
-	copy_from += copy_len;
+	copy_to += copy_len; copy_from += copy_len;
 
-	/* 3. format codes */
-
-	/* 3.1. the number of format codes */
+	/* format code */
 	memcpy(&tmp2, copy_from, sizeof(int16));
-	num_formats = ntohs(tmp2);
+	copy_len = sizeof(int16);
+	tmp2 = num_formats = ntohs(tmp2);
 
-	if (num_formats == 0)
+	if (num_formats >= 1)
 	{
-		/*
-		 * If num_formats is 0, the original message has no parameters or the parameter formats are all text,
-		 * so we don't need additional format codes since timestamp parametes use text as its format.
-		 */
-		num_formats_new = 0;
-	}
-	else
-	{
-		/* If num formats is 1, this means the specified format code is applied for all original parameters,
-		 * so enlarge message length to specify format codes for each of original paramters. */
+		/* this means the specified format code is applied all original parameters */
 		if (num_formats == 1)
+		{
 			*len += (num_org_params - 1) * sizeof(int16);
+			tmp2 += num_org_params - 1;
+		}
 
 		/* enlarge message length for timestamp parameter's formats */
 		*len += message->num_tsparams * sizeof(int16);
-		num_formats_new = num_org_params + message->num_tsparams;
+		tmp2 += message->num_tsparams;
 	}
 
-	/* copy the number of format codes */
-	tmp2 = htons(num_formats_new);
-	copy_len = sizeof(int16);
-	memcpy(copy_to, &tmp2, copy_len);
-	copy_to += copy_len;
-	copy_from += copy_len;
+	tmp2 = htons(tmp2);
+	memcpy(copy_to, &tmp2, copy_len);	/* copy number of format codes */
+	copy_to += copy_len; copy_from += copy_len;
+
+	copy_len = num_formats * sizeof(int16);
+
+	memcpy(copy_to, copy_from, copy_len);		/* copy format codes */
+	copy_to += copy_len; copy_from += copy_len;
 
-	/* 3.2. the format codes */
 	if (num_formats >= 1)
 	{
-
-		/* If num_formats is 1, copy the specified format code as numbers of original parameters */
+		/* copy the specified format code as numbers of original parameters */
 		if (num_formats == 1)
 		{
-			int16	org_format_code;
-			memcpy(&org_format_code, copy_from, sizeof(int16));
-			copy_from += copy_len;
-
-			for (i = 0; i < num_org_params; i++)
-			{
-				memcpy(copy_to, &org_format_code, sizeof(int16));
-				copy_to += sizeof(int16);
-			}
-		}
-		/* otherwize, copy the original format codes as they are*/
-		else
-		{
-			copy_len = num_formats * sizeof(int16);
-			memcpy(copy_to, copy_from, copy_len);
-			copy_to += copy_len;
-			copy_from += copy_len;
+			memcpy(copy_to, copy_from, (num_org_params - 1) * sizeof(int16));
+			copy_to += (num_org_params - 1) * sizeof(int16);
 		}
 
 		/* set additional format codes to zero(text) */
@@ -958,8 +930,6 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 		copy_to += sizeof(int16) * message->num_tsparams;
 	}
 
-	/* 4. parameters */
-
 	/* num params */
 	memcpy(&tmp2, copy_from, sizeof(int16));
 	copy_len = sizeof(int16);
@@ -968,7 +938,7 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	memcpy(copy_to, &tmp2, sizeof(int16));
 	copy_to += copy_len; copy_from += copy_len;
 
-	/* original params */
+	/* params */
 	copy_len = 0;
 	for (i = 0; i < num_params; i++)
 	{
@@ -987,7 +957,6 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 	memcpy(copy_to, copy_from, copy_len);
 	copy_to += copy_len; copy_from += copy_len;
 
-	/* timestamp params */
 	tmp4 = htonl(ts_len);
 	for (i = 0; i < message->num_tsparams; i++)
 	{
@@ -997,24 +966,16 @@ bind_rewrite_timestamp(POOL_CONNECTION_POOL *backend,
 		copy_to += ts_len;
 	}
 
-	/* 5. result format code */
+	/* result format code */
 	memcpy(&tmp2, copy_from, sizeof(int16));
 	copy_len = sizeof(int16);
 	copy_len += sizeof(int16) * ntohs(tmp2);
 	memcpy(copy_to, copy_from, copy_len);
 
-#ifdef TIMESTAMPDEBUG
-	fprintf(stderr, "message length:%d\n", *len);
-	for(i = 0; i < *len; i++)
-	{
-		fprintf(stderr, "%02x ", new_msg[i]);
-	}
-	fprintf(stderr, "\n");
-#endif
-
 	return new_msg;
 }
 
+/* make A_Const of T_String from "SELECT <expression>"*/
 static A_Const *makeStringConstFromQuery(POOL_CONNECTION_POOL *backend, char *expression)
 {
 	A_Const *con;
@@ -1059,7 +1020,9 @@ static A_Const *makeStringConstFromQuery(POOL_CONNECTION_POOL *backend, char *ex
  * that could appear under it, but not other statement types.
  */
 bool
-			raw_expression_tree_walker(Node *node, bool (*walker) (), void *context)
+raw_expression_tree_walker(Node *node,
+						   bool (*walker) (),
+						   void *context)
 {
 	ListCell   *temp;
 
@@ -1094,6 +1057,8 @@ bool
 			break;
 		case T_RangeVar:
 			return walker(((RangeVar *) node)->alias, context);
+		case T_GroupingFunc:
+			return walker(((GroupingFunc *) node)->args, context);
 		case T_SubLink:
 			{
 				SubLink    *sublink = (SubLink *) node;
@@ -1170,6 +1135,9 @@ bool
 				if (walker(into->rel, context))
 					return true;
 				/* colNames, options are deemed uninteresting */
+				/* viewQuery should be null in raw parsetree, but check it */
+				if (walker(into->viewQuery, context))
+					return true;
 			}
 			break;
 		case T_List:
@@ -1179,6 +1147,58 @@ bool
 					return true;
 			}
 			break;
+		case T_InsertStmt:
+			{
+				InsertStmt *stmt = (InsertStmt *) node;
+
+				if (walker(stmt->relation, context))
+					return true;
+				if (walker(stmt->cols, context))
+					return true;
+				if (walker(stmt->selectStmt, context))
+					return true;
+				if (walker(stmt->onConflictClause, context))
+					return true;
+				if (walker(stmt->returningList, context))
+					return true;
+				if (walker(stmt->withClause, context))
+					return true;
+			}
+			break;
+		case T_DeleteStmt:
+			{
+				DeleteStmt *stmt = (DeleteStmt *) node;
+
+				if (walker(stmt->relation, context))
+					return true;
+				if (walker(stmt->usingClause, context))
+					return true;
+				if (walker(stmt->whereClause, context))
+					return true;
+				if (walker(stmt->returningList, context))
+					return true;
+				if (walker(stmt->withClause, context))
+					return true;
+			}
+			break;
+		case T_UpdateStmt:
+			{
+				UpdateStmt *stmt = (UpdateStmt *) node;
+
+				if (walker(stmt->relation, context))
+					return true;
+				if (walker(stmt->targetList, context))
+					return true;
+				if (walker(stmt->whereClause, context))
+					return true;
+				if (walker(stmt->fromClause, context))
+					return true;
+				if (walker(stmt->returningList, context))
+					return true;
+				if (walker(stmt->withClause, context))
+					return true;
+			}
+			break;
 		case T_SelectStmt:
 			{
 				SelectStmt *stmt = (SelectStmt *) node;
@@ -1199,8 +1219,6 @@ bool
 					return true;
 				if (walker(stmt->windowClause, context))
 					return true;
-				if (walker(stmt->withClause, context))
-					return true;
 				if (walker(stmt->valuesLists, context))
 					return true;
 				if (walker(stmt->sortClause, context))
@@ -1211,6 +1229,8 @@ bool
 					return true;
 				if (walker(stmt->lockingClause, context))
 					return true;
+				if (walker(stmt->withClause, context))
+					return true;
 				if (walker(stmt->larg, context))
 					return true;
 				if (walker(stmt->rarg, context))
@@ -1228,6 +1248,14 @@ bool
 				/* operator name is deemed uninteresting */
 			}
 			break;
+		case T_BoolExpr:
+			{
+				BoolExpr   *expr = (BoolExpr *) node;
+
+				if (walker(expr->args, context))
+					return true;
+			}
+			break;
 		case T_ColumnRef:
 			/* we assume the fields contain nothing interesting */
 			break;
@@ -1237,11 +1265,17 @@ bool
 
 				if (walker(fcall->args, context))
 					return true;
+				if (walker(fcall->agg_order, context))
+					return true;
+				if (walker(fcall->agg_filter, context))
+					return true;
 				if (walker(fcall->over, context))
 					return true;
 				/* function name is deemed uninteresting */
 			}
 			break;
+		case T_NamedArgExpr:
+			return walker(((NamedArgExpr *) node)->arg, context);
 		case T_A_Indices:
 			{
 				A_Indices  *indices = (A_Indices *) node;
@@ -1274,6 +1308,8 @@ bool
 					return true;
 			}
 			break;
+		case T_MultiAssignRef:
+			return walker(((MultiAssignRef *) node)->source, context);
 		case T_TypeCast:
 			{
 				TypeCast   *tc = (TypeCast *) node;
@@ -1284,6 +1320,8 @@ bool
 					return true;
 			}
 			break;
+		case T_CollateClause:
+			return walker(((CollateClause *) node)->arg, context);
 		case T_SortBy:
 			return walker(((SortBy *) node)->node, context);
 		case T_WindowDef:
@@ -1294,6 +1332,10 @@ bool
 					return true;
 				if (walker(wd->orderClause, context))
 					return true;
+				if (walker(wd->startOffset, context))
+					return true;
+				if (walker(wd->endOffset, context))
+					return true;
 			}
 			break;
 		case T_RangeSubselect:
@@ -1314,6 +1356,21 @@ bool
 					return true;
 				if (walker(rf->alias, context))
 					return true;
+				if (walker(rf->coldeflist, context))
+					return true;
+			}
+			break;
+		case T_RangeTableSample:
+			{
+				RangeTableSample *rts = (RangeTableSample *) node;
+
+				if (walker(rts->relation, context))
+					return true;
+				/* method name is deemed uninteresting */
+				if (walker(rts->args, context))
+					return true;
+				if (walker(rts->repeatable, context))
+					return true;
 			}
 			break;
 		case T_TypeName:
@@ -1335,9 +1392,13 @@ bool
 					return true;
 				if (walker(coldef->raw_default, context))
 					return true;
+				if (walker(coldef->collClause, context))
+					return true;
 				/* for now, constraints are ignored */
 			}
 			break;
+		case T_GroupingSet:
+			return walker(((GroupingSet *) node)->content, context);
 		case T_LockingClause:
 			return walker(((LockingClause *) node)->lockedRels, context);
 		case T_XmlSerialize:
@@ -1352,6 +1413,28 @@ bool
 			break;
 		case T_WithClause:
 			return walker(((WithClause *) node)->ctes, context);
+		case T_InferClause:
+			{
+				InferClause *stmt = (InferClause *) node;
+
+				if (walker(stmt->indexElems, context))
+					return true;
+				if (walker(stmt->whereClause, context))
+					return true;
+			}
+			break;
+		case T_OnConflictClause:
+			{
+				OnConflictClause *stmt = (OnConflictClause *) node;
+
+				if (walker(stmt->infer, context))
+					return true;
+				if (walker(stmt->targetList, context))
+					return true;
+				if (walker(stmt->whereClause, context))
+					return true;
+			}
+			break;
 		case T_CommonTableExpr:
 			return walker(((CommonTableExpr *) node)->ctequery, context);
 		default:
